from __future__ import annotations
from typing import List, Optional
from dataclasses import dataclass, field

from jigu.core.sdk import Coin, Coins, Timestamp
from jigu.core.event import EVENT_TYPES
from jigu.core.msg import StdMsg, MSG_TYPES
from jigu.util.serdes import JSONSerializable, JSONDeserializable
from jigu.util.validation import Schemas as S

from .account import PublicKey


@dataclass
class StdFee(JSONSerializable, JSONDeserializable):

    __schema__ = S.OBJECT(gas=S.STRING_INTEGER, amount=Coins.__schema__)

    gas: int = 0
    amount: Coins = field(default_factory=Coins)

    def __post_init__(self):
        self.amount = Coins(self.amount)

    def to_data(self) -> dict:
        return {"gas": str(self.gas), "amount": self.amount}

    @classmethod
    def make(cls, gas: int = 0, **denoms):
        amount = Coins()
        for denom in denoms:
            amount += Coin(denom, denoms[denom])
        return cls(gas=gas, amount=amount)

    @classmethod
    def from_data(cls, data: dict) -> StdFee:
        return cls(gas=int(data["gas"]), amount=Coins.from_data(data["amount"]))


@dataclass
class StdSignature(JSONSerializable, JSONDeserializable):

    __schema__ = S.OBJECT(signature=S.STRING, pub_key=PublicKey.__schema__)

    signature: str
    pub_key: PublicKey

    def __str__(self) -> str:
        return self.signature

    @classmethod
    def from_data(cls, data: dict) -> StdSignature:
        return cls(
            signature=data["signature"], pub_key=PublicKey.from_data(data["pub_key"])
        )


@dataclass
class StdSignMsg(JSONSerializable):

    # TODO: Add deserialization?

    chain_id: Optional[str] = None
    account_number: Optional[int] = None
    sequence: Optional[int] = None
    fee: Optional[StdFee] = None
    msgs: List[StdMsg] = field(default_factory=list)
    memo: str = ""

    def to_data(self) -> dict:
        d = dict(self.__dict__)
        d["account_number"] = str(self.account_number)
        d["sequence"] = str(self.sequence)
        return d


@dataclass
class StdTx(JSONSerializable, JSONDeserializable):

    __schema__ = S.OBJECT(
        type=S.STRING_WITH_PATTERN(r"^core/StdTx\Z"),
        value=S.OBJECT(
            fee=StdFee.__schema__,
            msg=S.ARRAY(S.ANY(*(mt.__schema__ for mt in MSG_TYPES.values()))),
            signatures=S.ARRAY(StdSignature.__schema__),
            memo=S.STRING,
        ),
    )

    # NOTE: msg is not plural, and is NOT a typo. This may change later for consistency.
    fee: Optional[StdFee] = None
    msg: List[StdMsg] = field(default_factory=list)
    signatures: List[StdSignature] = field(default_factory=list)
    memo: str = ""

    def to_data(self) -> dict:
        return {
            "type": "core/StdTx",
            "value": dict(self.__dict__),
        }

    @classmethod
    def from_data(cls, data: dict) -> StdTx:
        data = data["value"]
        fee = StdFee.from_data(data["fee"])
        # deserialize the messages
        msg = []
        for m in data["msg"]:
            msg_type = MSG_TYPES[m["type"]]
            msg.append(msg_type.from_data(m))
        signatures = [StdSignature.from_data(s) for s in data["signatures"]]
        return cls(fee=fee, msg=msg, signatures=signatures, memo=data["memo"])


@dataclass
class TxInfo(JSONDeserializable):

    # TODO: Add Schema
    __schema__ = {}

    # TODO: Break TxInfo into BroadcastResult

    height: Optional[int] = None
    txhash: Optional[str] = None
    logs: Optional[List[dict]] = None
    gas_wanted: Optional[int] = None
    gas_used: Optional[int] = None
    timestamp: Optional[Timestamp] = None
    tx: Optional[StdTx] = None

    @classmethod
    def from_data(cls, data: dict) -> TxInfo:
        height = int(data["height"]) if "height" in data else None
        txhash = data.get("txhash", None)
        gas_wanted = int(data["gas_wanted"]) if "gas_wanted" in data else None
        gas_used = int(data["gas_used"]) if "gas_used" in data else None
        tx = StdTx.from_data(data["tx"]) if "tx" in data else None
        _logs = data.get("logs", None)
        if _logs is None:
            logs = None
        else:
            logs = []
            for l in _logs:
                for e in l["events"]:
                    if e["type"] in EVENT_TYPES:
                        e_type = EVENT_TYPES[e["type"]]
                        logs.append(e_type.from_data(e))
                    else:
                        logs.append("e")
        timestamp = (
            Timestamp.from_data(data["timestamp"]) if "timestamp" in data else None
        )
        return cls(
            height=height,
            txhash=txhash,
            logs=logs,
            gas_wanted=gas_wanted,
            gas_used=gas_used,
            tx=tx,
            timestamp=timestamp,
        )
