from __future__ import annotations
from typing import List
from dataclasses import dataclass

from jigu.core.auth.transaction import TxInfo
from jigu.core.sdk import Timestamp
from jigu.util import hash_amino, pretty_repr
from jigu.util.serdes import JSONDeserializable
from jigu.util.validation import Schemas as S


@dataclass
class BlockHeader(JSONDeserializable):

    __schema__ = S.OBJECT(
        block_id=S.OBJ,
        header=S.OBJECT(
            chain_id=S.STRING,
            height=S.STRING_INTEGER,
            time=Timestamp.__schema__,
            num_txs=S.STRING_INTEGER,
            total_txs=S.STRING_INTEGER,
            last_commit_hash=S.STRING,
            data_hash=S.STRING,
            next_validators_hash=S.STRING,
            consensus_hash=S.STRING,
            app_hash=S.STRING,
            proposer_address=S.STRING,
            last_block_id=S.OBJ,
        ),
    )

    chain_id: str
    height: int
    time: Timestamp
    num_txs: int
    total_txs: int
    hashes: dict
    proposer: str
    block_id: dict
    last_block_id: dict

    @classmethod
    def from_data(cls, data: dict) -> BlockHeader:
        header = data["header"]
        hashes = {
            "last_commit": header["last_commit_hash"],
            "data": header["data_hash"],
            "validators": header["validators_hash"],
            "next_validators": header["next_validators_hash"],
            "consensus": header["consensus_hash"],
            "app": header["app_hash"],
            "last_results": header["last_results_hash"],
            "evidence": header["evidence_hash"],
        }
        return cls(
            chain_id=header["chain_id"],
            height=int(header["height"]),
            time=Timestamp.from_data(header["time"]),
            num_txs=int(header["num_txs"]),
            total_txs=int(header["total_txs"]),
            hashes=hashes,
            proposer=header["proposer_address"],
            block_id=data["block_id"],
            last_block_id=header["last_block_id"],
        )


@dataclass
class Block(JSONDeserializable):

    __schema__ = S.OBJECT(
        block_meta=BlockHeader.__schema__,
        block=S.OBJECT(data=S.OBJECT(txs=S.OPTIONAL(S.ARRAY(S.STRING)))),
    )

    header: BlockHeader
    txs: List[TxInfo]

    @classmethod
    def from_data(cls, data: dict, terra) -> Block:
        header = BlockHeader.from_data(data["block_meta"])
        transactions = data["block"]["data"]["txs"] or []
        txs = []
        for tx in transactions:
            txs.append(terra.get_tx_info(hash_amino(tx)))  # populate each txn
        return cls(header=header, txs=txs)
