from __future__ import annotations
from dataclasses import dataclass
from typing import Dict, Iterable, List, Union, Callable
from decimal import Decimal

from jigu.error import DenomIncompatibleError
from jigu.util.serdes import JSONDeserializable, JSONSerializable
from jigu.util.validation import Schemas as S

from .dec import Dec


@dataclass(frozen=True)
class Coin(JSONSerializable, JSONDeserializable):

    __schema__ = S.OBJECT(
        denom=S.STRING, amount=S.ANY(S.STRING_INTEGER, Dec.__schema__)
    )

    denom: str
    amount: Union[int, Decimal, Dec, float, str]  # all get converted to int or Dec

    def __post_init__(self):
        s = self
        if (
            isinstance(s.amount, Dec)
            or isinstance(s.amount, Decimal)
            or isinstance(s.amount, float)
            or (
                isinstance(s.amount, str)
                and "." in str(s.amount)
                or "E" in str(s.amount)
            )
        ):
            object.__setattr__(s, "amount", Dec(s.amount))
        else:
            try:
                object.__setattr__(s, "amount", int(s.amount))
            except ValueError:
                raise ValueError(f"unacceptable value for Coin amount: {s.amount}")

    def __repr__(self) -> str:
        return f"Coin('{self.denom}', {repr(self.amount)})"

    def __str__(self) -> str:
        return f"{self.amount}{self.denom}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Coin):
            return self.denom == other.denom and self.amount == other.amount
        else:
            return False

    def __add__(self, other: Coin) -> Union[Coin, Coins]:
        if isinstance(other, Coins):
            return other + self
        if not isinstance(other, Coin):
            raise TypeError(
                f"unsupported operand types for +: 'Coin' and '{type(other)}'"
            )
        if self.denom != other.denom:
            raise DenomIncompatibleError(
                f"unsupported Coin denoms for +: '{self.denom}' and '{other.denom}'"
            )
        cast = int
        if isinstance(self.amount, Dec) or isinstance(other.amount, Dec):
            cast = Dec
        return Coin(denom=self.denom, amount=cast(self.amount + other.amount))

    def __sub__(self, other: Coin) -> Coin:
        if not isinstance(other, Coin):
            raise TypeError(
                f"unsupported operand types for -: 'Coin' and '{type(other)}'"
            )
        if self.denom != other.denom:
            raise DenomIncompatibleError(
                f"unsupported Coin denoms for -: '{self.denom}' and '{other.denom}'"
            )
        cast = int
        if isinstance(self.amount, Dec) or isinstance(other.amount, Dec):
            cast = Dec
        return Coin(denom=self.denom, amount=cast(self.amount - other.amount))

    def __mul__(self, other) -> Coin:
        return Coin(denom=self.denom, amount=(self.amount * other))

    def __rmul__(self, other) -> Coin:
        return self * other

    def __truediv__(self, other) -> Coin:
        cast = type(self.amount)
        return Coin(denom=self.denom, amount=cast(self.amount / other))

    def __floordiv__(self, other) -> Coin:
        cast = type(self.amount)
        return Coin(denom=self.denom, amount=cast(self.amount // other))

    def __neg__(self):
        return Coin(denom=self.denom, amount=(-self.amount))

    def __abs__(self):
        return Coin(denom=self.denom, amount=abs(self.amount))

    def __pos__(self):
        return abs(self)

    def __lt__(self, other: Coin) -> bool:
        if self.denom != other.denom:
            raise DenomIncompatibleError(
                f"incompatible Coin denoms for <: '{self.denom}' and '{other.denom}'"
            )
        return self.amount < other.amount

    def __gt__(self, other: Coin) -> bool:
        if self.denom != other.denom:
            raise DenomIncompatibleError(
                f"incompatible Coin denoms for >: '{self.denom}' and '{other.denom}'"
            )
        return self.amount > other.amount

    def __ge__(self, other: Coin) -> bool:
        return self > other or self == other

    def __le__(self, other: Coin) -> bool:
        return self < other or self == other

    def to_data(self) -> dict:
        return {"denom": str(self.denom), "amount": str(self.amount)}

    @classmethod
    def from_data(cls, data: Dict[str, str]) -> Coin:
        return cls(denom=data["denom"], amount=data["amount"])


class Coins(JSONSerializable, JSONDeserializable):

    __schema__ = S.ARRAY(Coin.__schema__)

    def __init__(self, coins: Iterable[Coin] = None, **denoms):
        if coins is None:
            coins = []
        coins = list(coins) + [Coin(d, a) for d, a in denoms.items()]
        self._cd = dict()
        for coin in list(coins):
            if self._cd.get(coin.denom, None) is None:
                self._cd[coin.denom] = Coin(coin.denom, coin.amount)
            else:
                self._cd[coin.denom] = coin + self._cd[coin.denom]

    @classmethod
    def make(cls, **denoms):
        """Convenience method for creating a Coins object easily with denoms
        in keyword arguments.
        """
        return cls(Coin(denom, amount) for denom, amount in denoms.items())

    def __repr__(self) -> str:
        return f"Coins({repr(self.coins)})"

    def __str__(self) -> str:
        return ", ".join(str(coin) for coin in self.coins)

    def to_data(self) -> List[Dict[str, str]]:
        return [coin.to_data() for coin in self.coins]

    def __add__(self, other: Union[Coin, Coins]) -> Coins:
        if isinstance(other, Coins):
            return Coins(self.coins + other.coins)
        elif isinstance(other, Coin):
            return Coins(self.coins + [other])
        else:
            raise TypeError(
                f"unsupported operand types for +: 'Coins' and '{type(other)}'"
            )

    def __mul__(self, other: Union[int, float, Decimal, Dec]) -> Coins:
        return Coins([coin * other for coin in self.coins])

    def __rmul__(self, other) -> Coins:
        return self * other

    def __truediv__(self, other: Union[int, float, Decimal, Dec]) -> Coins:
        return Coins([coin / other for coin in self.coins])

    def __floordiv__(self, other: Union[int, float, Decimal, Dec]) -> Coins:
        return Coins([coin / other for coin in self.coins])

    def __eq__(self, other: object) -> Bool:
        if isinstance(other, Coins):
            return JSONSerializable.__eq__(self, other)
        elif isinstance(other, list):
            try:
                return JSONSerializable.__eq__(self, Coins(other))
            except AttributeError:
                return False
        else:
            return False

    @classmethod
    def from_data(cls, data: List[Dict[str, str]]) -> Coins:
        coins = map(Coin.from_data, data)
        return cls(coins)

    @property
    def denoms(self) -> List[str]:
        return sorted(list(self._cd.keys()))

    @property
    def coins(self) -> List[Coin]:
        return sorted(list(self._cd.values()), key=lambda c: c.denom)

    def filter(self, predicate: Callable[[Coin], bool]) -> Coins:
        return Coins([c for c in self.coins if predicate(c)])

    def __iter__(self):
        return iter(self.coins)

    def __getitem__(self, denom: str) -> Coin:
        return self._cd[denom]

    def __contains__(self, denom: str) -> bool:
        return denom in self._cd
