from __future__ import annotations
from datetime import datetime
from dataclasses import dataclass
from decimal import Decimal
import re

from jigu.util.serdes import JSONDeserializable, JSONSerializable
from jigu.util.validation import Schemas as S


@dataclass(repr=False, frozen=True)
class Timestamp(JSONSerializable, JSONDeserializable):

    __schema__ = S.STRING_WITH_PATTERN(
        r"^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})(\.\d+)?Z\Z"
    )

    year: int
    month: int
    day: int
    hour: int
    minute: int
    second: int
    ns: int = 0
    precision: int = 0

    @property
    def microsecond(self):
        return self.ns // 1000

    @property
    def nanosecond(self):
        return int(self.ns % 1000)

    @property
    def datetime(self):
        return datetime(
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
            self.microsecond,
        )

    def __str__(self) -> str:
        Y = str(self.year).rjust(4, "0")
        M = str(self.month).rjust(2, "0")
        D = str(self.day).rjust(2, "0")
        h = str(self.hour).rjust(2, "0")
        m = str(self.minute).rjust(2, "0")
        s = str(self.second).rjust(2, "0")
        r = f"{Y}-{M}-{D}T{h}:{m}:{s}"
        if self.precision > 0:
            r += "."
            r += str(self.microsecond * 1000 + self.nanosecond).rjust(9, "0")[
                : self.precision
            ]
        return r + "Z"

    def __post_init__(self):
        object.__setattr__(self, "precision", min(9, self.precision))

    def __repr__(self) -> str:
        s = self
        p = f", {s.precision}" if s.precision != 0 else ""
        return f"Timestamp({s.year}, {s.month}, {s.day}, {s.hour}, {s.minute}, {s.second}, {s.ns}{p})"

    def to_data(self) -> str:
        return str(self)

    @classmethod
    def from_datetime(cls, dt: datetime) -> Timestamp:
        return cls(
            year=dt.year,
            month=dt.month,
            day=dt.day,
            hour=dt.hour,
            minute=dt.minute,
            second=dt.second,
            ns=dt.microsecond * 1000,
            precision=6 if dt.microsecond > 0 else 0,
        )

    @classmethod
    def from_data(cls, data: str) -> Timestamp:
        m = re.match(cls.__schema__["pattern"], data)
        year, month, day, hour, minute, second = [int(m.group(x)) for x in range(1, 7)]
        if m.group(7) is not None:
            # keep only nanosecond precision
            ns = int(Decimal(m.group(7)) * 1_000_000_000)
            precision = min(9, len(m.group(7)) - 1)
        else:
            ns = 0
            precision = 0
        return cls(year, month, day, hour, minute, second, ns, precision)
