from __future__ import annotations
import hashlib
from mnemonic import Mnemonic
from ecdsa import SECP256k1, SigningKey
from ecdsa.util import sigencode_string_canonize

from jigu.key import Key, derive_root, derive_child


class MnemonicKey(Key):
    """Implements Key interface with 24-word mnemonic. This implementation exposes the private
    key inside the application, and may not be suited for situations that demand high security."""

    def __init__(self, mnemonic: str, account: int = 0, index: int = 0):
        self.mnemonic = mnemonic
        seed = Mnemonic("english").to_seed(self.mnemonic)
        root = derive_root(seed)
        child = derive_child(root, account, index)
        self.account = account
        self.index = index
        self._private_key = child.PrivateKey()
        self._public_key = child.PublicKey()

    @property
    def public_key(self) -> bytes:
        return self._public_key

    @classmethod
    def generate(cls, account: int = 0, index: int = 0) -> MnemonicKey:
        """Generate a random Mnemonic, with the private / public key pair with the BIP44 HD Path
        located at 44'/330'/<account>'/0/<index> (LUNA)."""
        return cls(
            mnemonic=Mnemonic("english").generate(256), account=account, index=index
        )

    def sign(self, payload: str) -> bytes:
        """Sign a payload. Uses ecdsa curves, SECP256k1 by default."""
        sk = SigningKey.from_string(self._private_key, curve=SECP256k1)
        return sk.sign_deterministic(
            payload.encode(),
            hashfunc=hashlib.sha256,
            sigencode=sigencode_string_canonize,
        )
