from __future__ import annotations
from typing import Optional, List, Union

from cached_property import cached_property

import jigu.query
import jigu.terra
from jigu.core import (
    Validator,
    ValAddress,
    AccAddress,
    Coin,
    Coins,
    ExchangeRateVote,
    ExchangeRatePrevote,
    UnbondingDelegation,
    Delegation,
)


class ValidatorQuery(object):
    def __init__(self, terra: jigu.terra.Terra, val_address: ValAddress):
        self.terra = terra
        self.val_address = ValAddress(val_address)

    def __str__(self):
        return self.val_address

    def __repr__(self):
        return f"ValidatorQuery({self.val_address!r}) -> {self.terra}"

    # Staking
    def info(self) -> Validator:
        return self.terra.staking.val_info(self.val_address)

    @cached_property
    def account(self):
        return self.terra.account(self.val_address.acc_address)

    def delegations(self, delegator: Optional[AccAddress] = None) -> List[Delegation]:
        return self.terra.staking.delegations(
            validator=self.val_address, delegator=delegator
        )

    def unbonding_delegations(
        self, delegator: Optional[AccAddress] = None
    ) -> List[UnbondingDelegation]:
        return self.terra.staking.unbonding_delegations(
            validator=self.val_address, delegator=delegator
        )

    def redelegations(self):
        return self.terra.staking.redelegations(validator=self.val_address)

    # Distribution
    def rewards(
        self, denom: Optional[str] = None, include_delegators: Optional[bool] = False
    ) -> Union[Coin, Coins]:
        if include_delegators:
            rewards = self.terra.distribution.get_val_fee_rewards(self.val_address)
        else:
            rewards = self.terra.distribution.get_val_rewards(self.val_address)
        if denom is None:
            return rewards
        else:
            return rewards[denom]

    def reward_info(self, key: Optional[str] = None):
        return self.terra.distribution.get_val_reward_info(self.val_address, key)

    # Oracle
    def feeder(self) -> jigu.query.AccountQuery:
        return self.terra.account(
            self.terra.oracle.get_feeder_address(validator=self.val_address)
        )

    def misses(self) -> int:
        return self.terra.oracle.get_misses(validator=self.val_address)

    def votes(self, denom: Optional[str] = None) -> List[ExchangeRateVote]:
        return self.terra.oracle.votes(validator=self.val_address, denom=denom)

    def prevotes(self, denom: Optional[str] = None) -> List[ExchangeRatePrevote]:
        return self.terra.oracle.prevotes(validator=self.val_address, denom=denom)
