from __future__ import annotations
from typing import Optional, List, Union, Type
from contextlib import contextmanager

import jigu.terra
from jigu.core import UnsignedTx, Fee, StdTx, StdMsg
from jigu.key import Key
from jigu.query.account import AccountQuery


class Wallet(AccountQuery):
    """A Wallet is an augmented AccountQuery, and provides `chain_id`, `account_number`,
    and `sequence` information for signing transactions, which is performed by a Key."""

    def __init__(self, terra: jigu.terra.Terra, key: Type[Key]):
        AccountQuery.__init__(self, terra, key.acc_address)
        self.key = key
        self.terra = terra
        self._account_number = 0
        self._manual_sequence = False
        self._sequence = None

    def __repr__(self):
        return f"Wallet<{self.address}> -> {self.terra}"

    @property
    def account_number(self):
        if self._account_number == 0:
            self._account_number = self.info().account_number
        return self._account_number

    @property
    def sequence(self) -> int:
        return self.info().sequence

    def create_tx(
        self,
        msgs: Union[StdMsg, List[StdMsg]],
        fee: Optional[Fee] = None,
        memo: str = "",
    ) -> UnsignedTx:
        if type(msgs) != list:
            msgs = [msgs]  # singular
        if not fee:
            # estimate our fee if fee not supplied
            tx = StdTx(msg=msgs, memo=memo)
            fee = self.terra.estimate_tx_fee(tx)
        if self._manual_sequence:
            sequence = self._sequence
            self._sequence += 1
        else:
            sequence = self.sequence
        return UnsignedTx(
            chain_id=self.terra.chain_id,
            account_number=self.account_number,
            sequence=sequence,
            fee=fee,
            msgs=msgs,
            memo=memo,
        )

    def create_and_sign_tx(
        self,
        msgs: Union[StdMsg, List[StdMsg]],
        fee: Optional[Fee] = None,
        memo: str = "",
    ):
        return self.key.sign_tx(self.create_tx(msgs, fee, memo))

    @contextmanager
    def manual(self) -> Wallet:
        self._sequence = self.sequence
        self._manual_sequence = True
        yield self
        self._manual_sequence = False
