from typing import Any

import abc
import json

import fastjsonschema

_cached_schemas = dict()


def try_to_dict(o):
    try:
        return o.to_data()
    except AttributeError:
        return None


def serialize_to_json(item: Any, sort: bool = False, debug: bool = False) -> str:
    """This is used in case the serialization strategy must be used outside of
    JSONSerializable
    """
    return json.dumps(
        item,
        indent=2 if debug else None,
        sort_keys=sort,
        separators=(",", ":") if not debug else None,
        default=try_to_dict,
    )


class JSONSerializable(object):
    def to_data(self) -> Any:
        return dict(self.__dict__)  # copy instead of changing object

    def to_json(self, sort: bool = False, debug: bool = False) -> str:
        return serialize_to_json(self.to_data(), sort=sort, debug=debug)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, JSONSerializable):
            return self.to_data() == other.to_data()
        else:
            return self.to_data() == other


class JSONDeserializable(object, metaclass=abc.ABCMeta):
    """Can deserialize back into an Jigu object from JSON-friendly Python object.
    """

    __schema__ = {}

    @property
    @abc.abstractmethod
    def __schema__(self) -> dict:
        """Deserializable object should provide a Draft 7 compliant JSONSchema to
        validate inputs.
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_data(cls, data):
        """JSONDeserializable must implement from_dict()
        """
        raise NotImplementedError

    @classmethod
    def deserialize(cls, data, *args, **kwargs):
        """Applies JSON Schema-validation checks before attempting to recreate a
        Python object. Calls `from_dict()` internally.
        """

        if cls not in _cached_schemas:
            _cached_schemas[cls] = fastjsonschema.compile(cls.__schema__)
        _cached_schemas[cls](data)  # validate
        return cls.from_data(data, *args, **kwargs)
